require( "internal.lua" )

class( "Heap" )

function Heap:__init(...)
	self.data = {}--Heap:_parse_args( table.pack( ... ) )
	self.Count = 0
	self.Compare = function(A,B) return A < B end
end

function Heap:__tostring( )
	return string.format( "Heap( size=%d )", self.Count )
end

function Heap:Add( item )
	local heap_item = { HeapIndex=self.Count, Item=item }
	self.data[self.Count] = heap_item
	self:SortUp( heap_item )
	self.Count = self.Count + 1
end

function Heap:_default_compare( A, B )
	return A < B
end

function Heap:SortUp( heap_item )
	while true do
		local parent_index = math.floor( ( ( heap_item.HeapIndex - 1 ) / 2 ) + 0.5 )
		local parent_heap_item = self.data[parent_index]
		if self.Compare( heap_item.Item, parent_heap_item.Item ) then
			self:Swap( heap_item, parent_heap_item )
		else
			break
		end
	end
end

function Heap:SortDown( heap_item )
	while true do
		local child_index_left = math.floor( heap_item.HeapIndex * 2 + 1 + 0.5 )
		local child_index_right = math.floor( heap_item.HeapIndex * 2 + 2  + 0.5 )
		local swap_index = 0
		if child_index_left < self.Count then
			swap_index = child_index_left
			if child_index_right < self.Count then
				if self.Compare(self.data[child_index_left].Item, self.data[child_index_right].Item ) then
					swap_index = child_index_right
				end
			end
			if self.Compare( heap_item.Item, self.data[swap_index].Item ) then
				self:Swap( heap_item, self.data[swap_index] )
			else
				return
			end
		else
			return
		end
	end
end

function Heap:Swap( heap_item_A, heap_item_B )
	self.data[heap_item_A.HeapIndex] = heap_item_B
	self.data[heap_item_B.HeapIndex] = heap_item_A
	local item_a_index = heap_item_A.HeapIndex
	heap_item_A.HeapIndex = heap_item_B.HeapIndex
	heap_item_B.HeapIndex = item_a_index
end

function Heap:RemoveFirst()
	local first_heap_item = self.data[0]
	self.Count = self.Count - 1
	self.data[0] = self.data[self.Count]
	self.data[0].HeapIndex = 0
	self:SortDown( self.data[0] )
	self.data[self.Count] = nil
	return first_heap_item.Item
end

function Heap:Size()
	return self.Count
end

function Heap:PeekFirst()
	return self.data[0].Item
end

function Heap:SetCompareFunction( f )
	self.Compare = f
end

function Heap:Clear( )
	for i=0,#self.data do
		print( i )
		self.data[i] = nil
	end
	self.Count = 0
end

--[[
Example with custom compare function for a custom item

class( "CustomItem" )
function CustomItem:__init( char, score )
	self.char = char
	self.score = score
end

h = Heap()
h:SetCompareFunction( function(A,B) return A.score < B.score end )
h:Add( CustomItem( 'a', 1 ) )
h:Add( CustomItem( 'b', 2 ) )
h:Add( CustomItem( 'c', 3 ) )
print( h:RemoveFirst().char )
]]
