require( "internal.lua" )

class( "List" )

function List._parse_args( args )
	local _l = #args
	if _l == 0 then
		return {}
	elseif _l == 1 then
		local _t = __type( args[1] )
		if _t == "table" then
			return args[1]
		elseif _t == "Queue" or _t == "Stack" then
			return args[1].data
		end
	end
	error( "Cannot Initialize List; Check arguments: "..type( args ) )
end

function List:__init( ... )
	self.data = List._parse_args( table.pack( ... ) )
end

function List:__tostring( )
	return string.format( "List( size=%d )", #self.data )
end

function List:Size( )
	return #self.data
end

function List:Iter( )
	local i = 0
	local n = #self.data
	return function( )
		i = i + 1
		if i <= n then return self.data[i] end
	end
end

function List:PushBack( value )
	table.insert( self.data, value )
end

function List:PushFront( value )
	table.insert( self.data, 1, value )
end

function List:PopFront()
	if #self.data == 0 then
		return nil
	end
	return table.remove( self.data, 1 )
end

function List:PopBack()
	return table.remove( self.data )
end

function List:Insert( index, value )
	table.insert( self.data, value, index )
end

function List:Remove( index )
	return table.remove( self.data, index )
end

function List:Erase( value )
	for i, v in ipairs( self.data ) do
		if v == value then
			return table.remove( self.data, i )
		end
	end
end

function List:Clear( )
	for i,_ in ipairs( self.data ) do
		self.data[i] = nil
	end
end
