require( "internal.lua" )

class( "Matrix" )

function Matrix._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Matrix" then
			return { a.m11, a.m12, a.m13, a.m14, a.m21, a.m22, a.m23, a.m24, a.m31, a.m32, a.m33, a.m34, a.m41, a.m42, a.m43, a.m44 }
		elseif __type( a ) == "table" then
			return Matrix:_parse_args( a )
		end
		error( "Cannot Initialize Matrix; Check arguments: "..type( a ) )
	elseif _l == 12 then
		local a = args
		return { a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9], a[10], a[11], a[12], 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 16 then
		local a = args
		return { a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9], a[10], a[11], a[12], a[13], a[14], a[15], a[16] }
	end
	error( "Cannot Initialize Matrix; Check arguments: "..type( args ) )
end

function Matrix:__init( ... )
	local args = table.pack( ... )
	if #args ~= 1 or args[1] ~= false then
		local v = Matrix._parse_args( args )
		self.m11, self.m12, self.m13, self.m14 = v[1], 	v[2], 	v[3], 	v[4]
		self.m21, self.m22, self.m23, self.m24 = v[5], 	v[6], 	v[7], 	v[8]
		self.m31, self.m32, self.m33, self.m34 = v[9], 	v[10], 	v[11], 	v[12]
		self.m41, self.m42, self.m43, self.m44 = v[13], v[14], 	v[15], 	v[16]
	end
end

function Matrix:__tostring( )
	return string.format( "Matrix( %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f )",
								self.m11, self.m12, self.m13, self.m14,
								self.m21, self.m22, self.m23, self.m24,
								self.m31, self.m32, self.m33, self.m34,
								self.m41, self.m42, self.m43, self.m44 )
end

function Matrix:__mul( o )
	return Matrix(	o.m11 * self.m11 + o.m21 * self.m12 + o.m31 * self.m13 + o.m41 * self.m14 ,
					o.m12 * self.m11 + o.m22 * self.m12 + o.m32 * self.m13 + o.m42 * self.m14 ,
					o.m13 * self.m11 + o.m23 * self.m12 + o.m33 * self.m13 + o.m43 * self.m14 ,
					o.m14 * self.m11 + o.m24 * self.m12 + o.m34 * self.m13 + o.m44 * self.m14 ,
					o.m11 * self.m21 + o.m21 * self.m22 + o.m31 * self.m23 + o.m41 * self.m24 ,
					o.m12 * self.m21 + o.m22 * self.m22 + o.m32 * self.m23 + o.m42 * self.m24 ,
					o.m13 * self.m21 + o.m23 * self.m22 + o.m33 * self.m23 + o.m43 * self.m24 ,
					o.m14 * self.m21 + o.m24 * self.m22 + o.m34 * self.m23 + o.m44 * self.m24 ,
					o.m11 * self.m31 + o.m21 * self.m32 + o.m31 * self.m33 + o.m41 * self.m34 ,
					o.m12 * self.m31 + o.m22 * self.m32 + o.m32 * self.m33 + o.m42 * self.m34 ,
					o.m13 * self.m31 + o.m23 * self.m32 + o.m33 * self.m33 + o.m43 * self.m34 ,
					o.m14 * self.m31 + o.m24 * self.m32 + o.m34 * self.m33 + o.m44 * self.m34 ,
					o.m11 * self.m41 + o.m21 * self.m42 + o.m31 * self.m43 + o.m41 * self.m44 ,
					o.m12 * self.m41 + o.m22 * self.m42 + o.m32 * self.m43 + o.m42 * self.m44 ,
					o.m13 * self.m41 + o.m23 * self.m42 + o.m33 * self.m43 + o.m43 * self.m44 ,
					o.m14 * self.m41 + o.m24 * self.m42 + o.m34 * self.m43 + o.m44 * self.m44 )
end

function Matrix:_set( o )
	self.m11, self.m12, self.m13, self.m14 = o.m11, o.m12, o.m13, o.m14
	self.m21, self.m22, self.m23, self.m24 = o.m21, o.m22, o.m23, o.m24
	self.m31, self.m32, self.m33, self.m34 = o.m31, o.m32, o.m33, o.m34
	self.m41, self.m42, self.m43, self.m44 = o.m41, o.m42, o.m43, o.m44
end

function Matrix:SetIdentity()
	self.m11, self.m12, self.m13, self.m14 = 1.0, 0.0, 0.0, 0.0
	self.m21, self.m22, self.m23, self.m24 = 0.0, 1.0, 0.0, 0.0
	self.m31, self.m32, self.m33, self.m34 = 0.0, 0.0, 1.0, 0.0
	self.m41, self.m42, self.m43, self.m44 = 0.0, 0.0, 0.0, 1.0
end

function Matrix:SetTranslation( t )
	self.m11, self.m12, self.m13, self.m14 = 1.0, 0.0, 0.0, 0.0
	self.m21, self.m22, self.m23, self.m24 = 0.0, 1.0, 0.0, 0.0
	self.m31, self.m32, self.m33, self.m34 = 0.0, 0.0, 1.0, 0.0
	self.m41, self.m42, self.m43, self.m44 = t.x, t.y, t.z, 1.0
end

function Matrix:SetScaling( s )
	self.m11, self.m12, self.m13, self.m14 = s.x, 0.0, 0.0, 0.0
	self.m21, self.m22, self.m23, self.m24 = 0.0, s.y, 0.0, 0.0
	self.m31, self.m32, self.m33, self.m34 = 0.0, 0.0, s.z, 0.0
	self.m41, self.m42, self.m43, self.m44 = 0.0, 0.0,  0.0, 1.0
end

function Matrix:SetRotationX( angle )
	local cs = math.cos( angle )
	local sn = math.sin( angle )
	self.m11, self.m12, self.m13, self.m14 = 1.0, 0.0, 0.0, 0.0
	self.m21, self.m22, self.m23, self.m24 = 0.0,  cs,  sn, 0.0
	self.m31, self.m32, self.m33, self.m34 = 0.0, -sn,  cs, 0.0
	self.m41, self.m42, self.m43, self.m44 = 0.0, 0.0, 0.0, 1.0
end

function Matrix:SetRotationY( angle )
	local cs = math.cos( angle )
	local sn = math.sin( angle )
	self.m11, self.m12, self.m13, self.m14 = cs, 0.0, -sn, 0.0
	self.m21, self.m22, self.m23, self.m24 = 0.0, 1.0, 0.0, 0.0
	self.m31, self.m32, self.m33, self.m34 = sn, 0.0, cs, 0.0
	self.m41, self.m42, self.m43, self.m44 = 0.0, 0.0, 0.0, 1.0
end

function Matrix:SetRotationZ( angle )
	local cs = math.cos( angle )
	local sn = math.sin( angle )
	self.m11, self.m12, self.m13, self.m14 = cs, sn, 0.0, 0.0
	self.m21, self.m22, self.m23, self.m24 = -sn, cs, 0.0, 0.0
	self.m31, self.m32, self.m33, self.m34 = 0.0, 0.0, 1.0, 0.0
	self.m41, self.m42, self.m43, self.m44 = 0.0, 0.0, 0.0, 1.0
end

function Matrix:SetRotationEuler( euler, order )
	local rx = Matrix.RotationX( math.rad( euler.x ) )
	local ry = Matrix.RotationY( math.rad( euler.y ) )
	local rz = Matrix.RotationZ( math.rad( euler.z ) )
	if order == 'XYZ' then self:_set( ( rx * ry ) * rz ) end
	if order == 'XZY' then self:_set( ( rx * rz ) * ry ) end
	if order == 'YXZ' then self:_set( ( ry * rx ) * rz ) end
	if order == 'YZX' then self:_set( ( ry * rz ) * rx ) end
	if order == 'ZXY' then self:_set( ( rz * rx ) * ry ) end
	if order == 'ZYX' then self:_set( ( rz * ry ) * rx ) end
end

function Matrix:SetRotationAxisAngle( ax, angle )
	local cs = math.cos( angle )
	local cs1 = math.cos( 1.0 - cs )
	local sn = math.sin( angle )
	self.m11 = ax.x * ax.x * cs1 + cs
	self.m12 = ax.y * ax.x * cs1 + ax.z * sn
	self.m13 = ax.z * ax.x * cs1 - ax.y * sn
	self.m14 = 0.0

	self.m21 = ax.x * ax.y * cs1 - ax.z * sn
	self.m22 = ax.y * ax.y * cs1 + cs
	self.m23 = ax.z * ax.y * cs1 + ax.x * sn
	self.m24 = 0.0

	self.m31 = ax.x * ax.z * cs1 + ax.y * sn
	self.m32 = ax.y * ax.z * cs1 - ax.x * sn
	self.m33 = ax.z * ax.z * cs1 + cs
	self.m34 = 0.0

	self.m41 = 0.0
	self.m42 = 0.0
	self.m43 = 0.0
	self.m44 = 1.0
end

function Matrix:SetRotationQuaternion( q )
	local xx, xy, xz, xw = q.x * q.x ,  q.x * q.y , q.x * q.z , q.x * q.w
	local yy, yz, yw = q.y * q.y , q.y * q.z , q.y * q.w
	local zz, zw = q.z * q.z, q.z * q.w
	self.m11 = 1.0 - 2.0 * ( zz + yy )
	self.m12 = 2.0 * ( xy + zw )
	self.m13 = 2.0 * ( xz - yw )
	self.m14 = 0.0

	self.m21 = 2.0 * ( xy - zw )
	self.m22 = 1.0 - 2.0 * ( zz + xx )
	self.m23 = 2.0 * ( yz + xw )
	self.m24 = 0.0

	self.m31 = 2.0 * ( yw + xz )
	self.m32 = 2.0 * ( yz - xw )
	self.m33 = 1.0 - 2.0 * ( yy + xx )
	self.m34 = 0.0

	self.m41 = 0.0
	self.m42 = 0.0
	self.m43 = 0.0
	self.m44 = 1.0
end

function Matrix:Determinant()
  	return self.m14 * self.m23 * self.m32 * self.m41 - self.m13 * self.m24 * self.m32 * self.m41 - self.m14 * self.m22 * self.m33 * self.m41 + self.m12 * self.m24 * self.m33 * self.m41 +
		self.m13 * self.m22 * self.m34 * self.m41 - self.m12 * self.m23 * self.m34 * self.m41 - self.m14 * self.m23 * self.m31 * self.m42 + self.m13 * self.m24 * self.m31 * self.m42 +
		self.m14 * self.m21 * self.m33 * self.m42 - self.m11 * self.m24 * self.m33 * self.m42 - self.m13 * self.m21 * self.m34 * self.m42 + self.m11 * self.m23 * self.m34 * self.m42 +
		self.m14 * self.m22 * self.m31 * self.m43 - self.m12 * self.m24 * self.m31 * self.m43 - self.m14 * self.m21 * self.m32 * self.m43 + self.m11 * self.m24 * self.m32 * self.m43 +
		self.m12 * self.m21 * self.m34 * self.m43 - self.m11 * self.m22 * self.m34 * self.m43 - self.m13 * self.m22 * self.m31 * self.m44 + self.m12 * self.m23 * self.m31 * self.m44 +
		self.m13 * self.m21 * self.m32 * self.m44 - self.m11 * self.m23 * self.m32 * self.m44 - self.m12 * self.m21 * self.m33 * self.m44 + self.m11 * self.m22 * self.m33 * self.m44
end

function Matrix:GetTranslation()
	return Vector( self.m41, self.m42, self.m43 )
end

function Matrix:GetScaling()
  	return Vector( 	math.sqrt( self.m11 * self.m11 + self.m12 * self.m12 + self.m13 * self.m13 ),
					math.sqrt( self.m21 * self.m21 + self.m22 * self.m22 + self.m23 * self.m23 ),
					math.sqrt( self.m31 * self.m31 + self.m32 * self.m32 + self.m33 * self.m33 ) )
end

function Matrix:GetRotation()
	if self.m11 + self.m22 + self.m33 > 0 then
		local s = math.sqrt( 1.0 + self.m11 + self.m22 + self.m33 ) * 2.0
		return Quaternion( ( self.m23 - self.m32 ) / s , ( self.m31 - self.m13 ) / s , ( self.m12 - self.m21 ) / s , 0.25 * s )
	end
	if ( self.m11 > self.m22 ) and ( self.m11 > self.m33 ) then
		local s = math.sqrt( 1.0 + self.m11 - self.m22 - self.m33 ) * 2.0
		return Quaternion(  0.25 * s , ( self.m21 + self.m12 ) / s , ( self.m31 + self.m13 ) / s , ( self.m23 - self.m32 ) / s )
	end
	if self.m22 > self.m33 then
		local s = math.sqrt( 1.0 + self.m22 - self.m11 - self.m33 ) * 2.0
		return Quaternion( ( self.m21 + self.m12 ) / s , 0.25 * s , ( self.m32 + self.m23) / s , ( self.m31 - self.m13 ) / s )
	end
	local s = math.sqrt( 1.0 + self.m33 - self.m11 - self.m22 ) * 2.0
	return Quaternion( ( self.m31 + self.m13 ) / s , ( self.m32 + self.m23 ) / s , 0.25 * s , ( self.m12 - self.m21 ) / s )
end

function Matrix:Decompose()
	return self:GetTranslation(), self:GetRotation(), self:GetScaling()
end

function Matrix:Inverse()
	local result = Matrix()
	local det = self:Determinant( )
	if math.abs( det ) > 0.00001 then
		det = 1.0 / det
	else
		return result
  	end
	result.m11 = det * ( self.m23*self.m34*self.m42 - self.m24*self.m33*self.m42 + self.m24*self.m32*self.m43 - self.m22*self.m34*self.m43 - self.m23*self.m32*self.m44 + self.m22*self.m33*self.m44 )
	result.m12 = det * ( self.m14*self.m33*self.m42 - self.m13*self.m34*self.m42 - self.m14*self.m32*self.m43 + self.m12*self.m34*self.m43 + self.m13*self.m32*self.m44 - self.m12*self.m33*self.m44 )
	result.m13 = det * ( self.m13*self.m24*self.m42 - self.m14*self.m23*self.m42 + self.m14*self.m22*self.m43 - self.m12*self.m24*self.m43 - self.m13*self.m22*self.m44 + self.m12*self.m23*self.m44 )
	result.m14 = det * ( self.m14*self.m23*self.m32 - self.m13*self.m24*self.m32 - self.m14*self.m22*self.m33 + self.m12*self.m24*self.m33 + self.m13*self.m22*self.m34 - self.m12*self.m23*self.m34 )

	result.m21 = det * ( self.m24*self.m33*self.m41 - self.m23*self.m34*self.m41 - self.m24*self.m31*self.m43 + self.m21*self.m34*self.m43 + self.m23*self.m31*self.m44 - self.m21*self.m33*self.m44 )
	result.m22 = det * ( self.m13*self.m34*self.m41 - self.m14*self.m33*self.m41 + self.m14*self.m31*self.m43 - self.m11*self.m34*self.m43 - self.m13*self.m31*self.m44 + self.m11*self.m33*self.m44 )
	result.m23 = det * ( self.m14*self.m23*self.m41 - self.m13*self.m24*self.m41 - self.m14*self.m21*self.m43 + self.m11*self.m24*self.m43 + self.m13*self.m21*self.m44 - self.m11*self.m23*self.m44 )
	result.m24 = det * ( self.m13*self.m24*self.m31 - self.m14*self.m23*self.m31 + self.m14*self.m21*self.m33 - self.m11*self.m24*self.m33 - self.m13*self.m21*self.m34 + self.m11*self.m23*self.m34 )

	result.m31 = det * ( self.m22*self.m34*self.m41 - self.m24*self.m32*self.m41 + self.m24*self.m31*self.m42 - self.m21*self.m34*self.m42 - self.m22*self.m31*self.m44 + self.m21*self.m32*self.m44 )
	result.m32 = det * ( self.m14*self.m32*self.m41 - self.m12*self.m34*self.m41 - self.m14*self.m31*self.m42 + self.m11*self.m34*self.m42 + self.m12*self.m31*self.m44 - self.m11*self.m32*self.m44 )
	result.m33 = det * ( self.m12*self.m24*self.m41 - self.m14*self.m22*self.m41 + self.m14*self.m21*self.m42 - self.m11*self.m24*self.m42 - self.m12*self.m21*self.m44 + self.m11*self.m22*self.m44 )
	result.m34 = det * ( self.m14*self.m22*self.m31 - self.m12*self.m24*self.m31 - self.m14*self.m21*self.m32 + self.m11*self.m24*self.m32 + self.m12*self.m21*self.m34 - self.m11*self.m22*self.m34 )

	result.m41 = det * ( self.m23*self.m32*self.m41 - self.m22*self.m33*self.m41 - self.m23*self.m31*self.m42 + self.m21*self.m33*self.m42 + self.m22*self.m31*self.m43 - self.m21*self.m32*self.m43 )
	result.m42 = det * ( self.m12*self.m33*self.m41 - self.m13*self.m32*self.m41 + self.m13*self.m31*self.m42 - self.m11*self.m33*self.m42 - self.m12*self.m31*self.m43 + self.m11*self.m32*self.m43 )
	result.m43 = det * ( self.m13*self.m22*self.m41 - self.m12*self.m23*self.m41 - self.m13*self.m21*self.m42 + self.m11*self.m23*self.m42 + self.m12*self.m21*self.m43 - self.m11*self.m22*self.m43 )
	result.m44 = det * ( self.m12*self.m23*self.m31 - self.m13*self.m22*self.m31 + self.m13*self.m21*self.m32 - self.m11*self.m23*self.m32 - self.m12*self.m21*self.m33 + self.m11*self.m22*self.m33 )
end

function Matrix:Inversed()
	local m = Matrix( self )
	m:Inverse()
	return m
end

function Matrix:Transposed()
	local result = Matrix()
	result.m11 = self.m11
	result.m12 = self.m21
	result.m13 = self.m31
	result.m14 = self.m41

	result.m21 = self.m12
	result.m22 = self.m22
	result.m23 = self.m32
	result.m24 = self.m42

	result.m31 = self.m13
	result.m32 = self.m23
	result.m33 = self.m33
	result.m34 = self.m43

	result.m41 = self.m14
	result.m42 = self.m24
	result.m43 = self.m34
	result.m44 = self.m44
	return result
end

function Matrix:RotateVector( v )
	return Vector(  v.x * self.m11 + v.y * self.m21 + v.z * self.m31,
					v.x * self.m12 + v.y * self.m22 + v.z * self.m32,
					v.x * self.m13 + v.y * self.m23 + v.z * self.m33 )
end

function Matrix:TransformVector( v )
	return Vector(  v.x * self.m11 + v.y * self.m21 + v.z * self.m31 + v.w * self.m41,
					v.x * self.m12 + v.y * self.m22 + v.z * self.m32 + v.w * self.m42,
					v.x * self.m13 + v.y * self.m23 + v.z * self.m33 + v.w * self.m43,
					v.x * self.m14 + v.y * self.m24 + v.z * self.m34 + v.w * self.m44 )
end

function Matrix:TransformVector3x3( v )
	return Vector(  v.x * self.m11 + v.y * self.m21 + v.z * self.m31,
					v.x * self.m12 + v.y * self.m22 + v.z * self.m32,
					v.x * self.m13 + v.y * self.m23 + v.z * self.m33 )
end

--
-- STATIC METHODS
--

function Matrix.Translation( t )
	local m = Matrix( false )
	m:SetTranslation( t )
	return m
end

function Matrix.Scaling( s )
	local m = Matrix( false )
	m:SetScaling( s )
	return m
end

function Matrix.RotationEuler( euler, axis )
	local m = Matrix( false )
	m:SetRotationEuler( euler, axis )
	return m
end

function Matrix.RotationAxisAngle( axis, angle )
	local m = Matrix( false )
	m:SetRotationAxisAngle( axis, angle )
	return m
end

function Matrix.RotationX( angle )
	local m = Matrix( false )
	m:SetRotationX( angle )
	return m
end

function Matrix.RotationY( angle )
	local m = Matrix( false )
	m:SetRotationY( angle )
	return m
end

function Matrix.RotationZ( angle )
	local m = Matrix( false )
	m:SetRotationZ( angle )
	return m
end
