require( "internal.lua" )

class( "Ray" )

function Ray._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { Vector(), Vector( 0.0, 1.0, 0.0 ), 1.0 }
	elseif _l == 1 and __type( args[1] ) == "Ray" then
		return args[1].start, args[1].direction, args[1].length
	elseif _l == 3 then
		return { args[1], args[2], args[3] }
	end
	error( "Cannot Initialize Ray; Check arguments: "..type( args ) )
end

function Ray:__init( ... )
	local args = Ray._parse_args( table.pack( ... ) )
	self.start = args[1]
	self.direction = args[2]
	self.length = args[3]
end

function Ray:__eq( other )
	return self.start == other.start and self.direction == other.direction and self.length == other.length
end

function Ray:__tostring( )
	return string.format( "Ray( start="..self.start..", direction="..self.direction..", length="..self.length.." )" )
end

function Ray:Length( )
	return self.length
end

function Ray:SetLength( l )
	self.length = l
end

function Ray:SetStart( ... )
	self.start = Vector( table.pack( ... ) )
end

function Ray:Start()
	return self.start
end

function Ray:SetDirection( ... )
	self.direction = Vector( table.pack( ... ) ):Normalized()
end

function Ray:Direction()
	return self.direction
end

function Ray:End( )
	return self.start + self.direction * self.length
end

function Ray:SetEnd( ... )
	local _end = Vector( table.pack( ... ) )
	local _diff = _end - self.start
	self.length = _diff:Length()
	self.direction = _diff:Normalized()
end

function Ray:Lerp( other, t )
	local s = self.start:Lerp( other.start, t )
	local d = self.direction:Lerp( other.direction, t )
	local l = lerp( self.length, other.length, t )
	return Ray( s, d, l )
end

function Ray.Bezier3( r1, r2, r3, t )
	local s = Vector.Bezier3( r1.start, r2.start, r3.start, t )
	local d = Vector.Bezier3( r1.direction, r2.direction, r3.direction, t )
	local l = bezier3( r1.length, r2.length, r3.length, t )
	return Ray( s, d, l )
end

function Ray.Bezier4( r1, r2, r3, r4, t )
	local s = Vector.Bezier4( r1.start, r2.start, r3.start, r4.start, t )
	local d = Vector.Bezier4( r1.direction, r2.direction, r3.direction, r4.direction, t )
	local l = bezier4( r1.length, r2.length, r3.length, r4.length, t )
	return Ray( s, d, l )
end
