function lerp( a, b, f )
	return b * f + a * ( 1.0 - f )
end

function clamp( a, b, f )
	if f < a then return a
	elseif f > b then return b
	end
	return f
end

function saturate( f )
	return clamp( 0.0, 1.0, f )
end

function smoothstep( edge0, edge1, x )
	local x = clamp( 0.0, 1.0, ( x - edge0 ) / ( edge1 - edge0 ) )
	return x * x * ( 3.0 - 2.0 * x )
end

function smootherstep( edge0, edge1, x )
	local x = clamp( 0.0, 1.0, ( x - edge0 ) / ( edge1 - edge0 ) )
	return x * x * x * ( x * ( x * 6.0 - 15.0 ) + 10.0 )
end

function gauss( mu, sigma, x )
	return math.exp( -pow( (x-mu), 2 ) / ( 2 * pow( sigma, 2 ) ) ) / math.sqrt( 2 * math.pi * pow( sigma, 2 ) )
end

function sign( x )
	if x < 0 then return -1
	elseif x == 0 then return 0
	end
	return 1
end

function bezier3( x0, x1, x2, t )
	local mt = 1.0 - t
	local mt2 = mt * mt
	return mt2 * x0 + 2.0 * mt * t * x1 + t * t * x2
end

function bezier4( p0, p1, p2, p3, t )
	local mt = 1.0 -t
	local mt2 = mt * mt
	local mt3 = mt2 * mt
	local t2 = t * t
	local t3 = t2 * t
	return mt3 * p0 + 3.0 * mt2 * t * p1 + 3.0 * mt * t2 * p2 + t3 * p3
end
